/*
 * Copyright (c) 1997 Guylhem Aznar <guylhem@oeil.qc.ca>
 * Copyright (c) 1994 Mike Finger <mfinger@mermaid.micro.umn.edu>
 *                    or <Mike_Finger@atk.com>
 * Copyright (c) 1994 Robert Nation
 * Copyright (c) 1994 Nobutaka Suzuki
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "../../configure.h"
#include <stdio.h>
#include <X11/Xlib.h>
#include "Colors.h"

#ifndef min
#define min(a,b) (((a)<(b)) ? (a) : (b))
#define max(a,b) (((a)>(b)) ? (a) : (b))
#endif

extern Display *dpy;
extern Window Root;

/****************************************************************************
  Loads a single color
*****************************************************************************/
Pixel
GetColor (char *name)
{
  XColor color;
  XWindowAttributes attributes;

  XGetWindowAttributes (dpy, Root, &attributes);
  color.pixel = 0;
  if (!XParseColor (dpy, attributes.colormap, name, &color))
    nocolor ("parse", name);
  else if (!XAllocColor (dpy, attributes.colormap, &color))
    nocolor ("alloc", name);
  return color.pixel;
}

/****************************************************************************
  This routine computes the hilight color from the background color
*****************************************************************************/
Pixel
GetHilite (Pixel background)
{
  XColor bg_color, white_p;
  XWindowAttributes attributes;

  XGetWindowAttributes (dpy, Root, &attributes);

  bg_color.pixel = background;
  XQueryColor (dpy, attributes.colormap, &bg_color);

  white_p.pixel = GetColor ("white");
  XQueryColor (dpy, attributes.colormap, &white_p);

  bg_color.red = max ((white_p.red / 5), bg_color.red);
  bg_color.green = max ((white_p.green / 5), bg_color.green);
  bg_color.blue = max ((white_p.blue / 5), bg_color.blue);

  bg_color.red = min (white_p.red, (bg_color.red * 140) / 100);
  bg_color.green = min (white_p.green, (bg_color.green * 140) / 100);
  bg_color.blue = min (white_p.blue, (bg_color.blue * 140) / 100);

  if (!XAllocColor (dpy, attributes.colormap, &bg_color))
    nocolor ("alloc hilight", "");

  return bg_color.pixel;
}

/****************************************************************************
  This routine computes the shadow color from the background color
*****************************************************************************/
Pixel
GetShadow (Pixel background)
{
  XColor bg_color;
  XWindowAttributes attributes;

  XGetWindowAttributes (dpy, Root, &attributes);

  bg_color.pixel = background;
  XQueryColor (dpy, attributes.colormap, &bg_color);

  bg_color.red = (unsigned short) ((bg_color.red * 50) / 100);
  bg_color.green = (unsigned short) ((bg_color.green * 50) / 100);
  bg_color.blue = (unsigned short) ((bg_color.blue * 50) / 100);

  if (!XAllocColor (dpy, attributes.colormap, &bg_color))
    nocolor ("alloc shadow", "");

  return bg_color.pixel;
}

void
nocolor (char *a, char *b)
{
  fprintf (stderr, "WinList: can't %s %s\n", a, b);
}
